#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 2013-2014 California Institute of Technology.
# License: 3-clause BSD.  The full license text is available at:
#  - http://trac.mystic.cacr.caltech.edu/project/pathos/browser/klepto/LICENSE

from klepto import lru_cache as memoize
from klepto.keymaps import hashmap
hasher = hashmap(algorithm='md5')

class Adder(object):
    """A simple class with a memoized method"""

    @memoize(keymap=hasher, ignore=('self','**'))
    def __call__(self, x, *args, **kwds):
        debug = kwds.get('debug', False)
        if debug:
            print ('debug:', x, args, kwds)
        return sum((x,)+args)
    add = __call__


add = Adder()
assert add(2,0) == 2
assert add(2,0,z=4) == 2          # cached (ignore z)
assert add(2,0,debug=False) == 2  # cached (ignore debug)
assert add(1,2,debug=False) == 3
assert add(1,2,debug=True) == 3   # cached (ignore debug)
assert add(4) == 4
assert add(x=4) == 4              # cached

plus = Adder()
assert plus(4,debug=True) == 4    # cached (ignore debug)
assert plus(2,0,3) == 5
assert plus(2,0,4) == 6

info = add.__call__.info()
assert info.hit == 5
assert info.miss == 5
cache = add.__call__.__cache__()
assert sorted(cache.values()) == [2,3,4,5,6]

# careful... remember we have self as first argument
key = add.__call__.key(add,2,0)
assert cache[key] == add.__call__.__wrapped__(add,2,0)
assert cache[key] == add.__call__.lookup(add,2,0)


######################################################
@memoize(keymap=hasher, ignore=('self','**'))
def _add(x, *args, **kwds):
    debug = kwds.get('debug', False)
    if debug:
        print ('debug:', x, args, kwds)
    return sum((x,)+args)

_add(2,0)
_add(2,0,z=4)
_add(2,0,debug=False)
_add(1,2,debug=False)
_add(1,2,debug=True)
_add(4)
_add(x=4)
_add(4,debug=True)
_add(2,0,3)
_add(2,0,4)

_cache =  _add.__cache__()
_func = _add.__wrapped__

# do a lookup
assert _add.lookup(2,0) == _func(2,0)

# generate the key, and do a look-up
key = _add.key(2,0)
assert _cache[key] == _func(2,0)

# look-up the key again, doing a little more work...
from klepto import keygen
lookup = keygen('self','**')(_func)
lookup.register(hasher)
key = lookup(2,0)
assert _cache[key] == _func(2,0)

# since we have the 'key lookup', let's play with lookup a bit
assert lookup.valid()
assert lookup.call() == _func(2,0)


######################################################
# more of the same...
from klepto import inf_cache

class Foo(object):
  @keygen('self')
  def bar(self, x,y):
    return x+y

fu = Foo()
assert fu.bar(1,2) == ('x', 1, 'y', 2)
assert Foo.bar(fu,1,2) == ('x', 1, 'y', 2)

class _Foo(object):
  @inf_cache(ignore='self')
  def bar(self, x,y):
    return x+y

_fu = _Foo()
_fu.bar(1,2)
_fu.bar(2,2)
_fu.bar(2,3)
_fu.bar(1,2)
assert len(_fu.bar.__cache__()) == 3
assert _fu.bar.__cache__()[_fu.bar.key(_fu,1,2)] == 3
assert _fu.bar.lookup(_fu,1,2) == 3


# EOF



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 2013-2014 California Institute of Technology.
# License: 3-clause BSD.  The full license text is available at:
#  - http://trac.mystic.cacr.caltech.edu/project/pathos/browser/klepto/LICENSE

from klepto.keymaps import *
h = hashmap(algorithm='md5')
p = picklemap(serializer='dill')
hp = p + h

try:
    import numpy as np
    x = np.arange(2000)
    y = x.copy()
    y[1000] = -1

    assert h(x) == h(y) # equal because repr for large np arrays uses '...'
    assert p(x) != p(y)
    assert hp(x) != hp(y)
except ImportError:
    print("to test big data, install numpy")


# EOF



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 2013-2014 California Institute of Technology.
# License: 3-clause BSD.  The full license text is available at:
#  - http://trac.mystic.cacr.caltech.edu/project/pathos/browser/klepto/LICENSE

from klepto.rounding import *

@deep_round(tol=1)
def add(x,y):
  return x+y

result = add(2.54, 5.47)
assert result == 8.0

# rounds each float, regardless of depth in an object
result = add([2.54, 5.47],['x','y'])
assert result == [2.5, 5.5, 'x', 'y']

# rounds each float, regardless of depth in an object
result = add([2.54, 5.47],['x',[8.99, 'y']])
assert result == [2.5, 5.5, 'x', [9.0, 'y']]

@simple_round(tol=1)
def add(x,y):
  return x+y

result = add(2.54, 5.47)
assert result == 8.0

# does not round elements of iterables, only rounds at the top-level
result = add([2.54, 5.47],['x','y'])
assert result == [2.54, 5.4699999999999998, 'x', 'y']

# does not round elements of iterables, only rounds at the top-level
result = add([2.54, 5.47],['x',[8.99, 'y']])
assert result == [2.54, 5.4699999999999998, 'x', [8.9900000000000002, 'y']]

@shallow_round(tol=1)
def add(x,y):
  return x+y

result = add(2.54, 5.47)
assert result == 8.0

# rounds each float, at the top-level or first-level of each object.
result = add([2.54, 5.47],['x','y'])
assert result == [2.5, 5.5, 'x', 'y']

# rounds each float, at the top-level or first-level of each object.
result = add([2.54, 5.47],['x',[8.99, 'y']])
assert result == [2.5, 5.5, 'x', [8.9900000000000002, 'y']]


# rounding integrated with key generation
from klepto import keygen, NULL

@keygen('x',2,tol=2)
def add(w,x,y,z):
    return x+y+z+w

assert add(1.11111,2.222222,3.333333,4.444444) == ('w', 1.11, 'x', NULL, 'y', NULL, 'z', 4.44)
assert add.call() == 11.111108999999999
assert add(1.11111,2.2229,100,4.447) == ('w', 1.11, 'x', NULL, 'y', NULL, 'z', 4.45)
assert add.call() == 107.78101
assert add(1.11111,100,100,4.441) == ('w', 1.11, 'x', NULL, 'y', NULL, 'z', 4.44)
assert add.call() == 205.55211



# EOF



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 2013-2014 California Institute of Technology.
# License: 3-clause BSD.  The full license text is available at:
#  - http://trac.mystic.cacr.caltech.edu/project/pathos/browser/klepto/LICENSE

import sys
from functools import partial
def foo(x,y,z,a=1,b=2):
    return x+y+z+a+b

from klepto import validate
p = partial(foo, 0,1)
try:
    res1 = p(1,2,3,4,b=5)
    res2 = Exception()
except:
    res1,res2 = sys.exc_info()[:2]
try:
    re_1 = validate(p,1,2,3,4,b=5)
    re_2 = Exception()
except:
    re_1,re_2 = sys.exc_info()[:2]
#print(res2)
#print(re_2)
assert res1 == re_1
if hex(sys.hexversion).startswith('0x207'):
    assert str(res2) == str(TypeError("foo() takes at most 5 arguments (7 given)"))
elif hex(sys.hexversion) < '0x30300f0': # all versions slightly different
    assert str(res2)[:20] == str(re_2)[:20]
else:
    pass # python 3.3 gives "foo() got multiple values for argument 'b'"

try:
    res1 = p()
    res2 = Exception()
except:
    res1,res2 = sys.exc_info()[:2]
try:
    re_1 = validate(p)
    re_2 = Exception()
except:
    re_1,re_2 = sys.exc_info()[:2]
#print(res2)
#print(re_2)
assert res1 == re_1
if hex(sys.hexversion).startswith('0x207'):
    assert str(res2) == str(TypeError("foo() takes at least 3 arguments (2 given)"))
elif hex(sys.hexversion) < '0x30300f0': # all versions slightly different
    assert str(res2)[:21] == str(re_2)[:21]
else:
    pass # python 3.3 gives "foo() missing 1 required positional argument 'z'"

try:
    res1 = p(1,2,3,4,r=5)
    res2 = Exception()
except:
    res1,res2 = sys.exc_info()[:2]
try:
    re_1 = validate(p,1,2,3,4,r=5)
    re_2 = Exception()
except:
    re_1,re_2 = sys.exc_info()[:2]
#print(res2)
#print(re_2)
assert res1 == re_1
if hex(sys.hexversion).startswith('0x207'):
    assert str(res2) == str(TypeError("foo() takes at most 5 arguments (7 given)"))
elif hex(sys.hexversion) < '0x30300f0': # all versions slightly different
    assert str(res2)[:20] == str(re_2)[:20]
else:
    pass # python 3.3 gives "foo() got unexpected keyword argument 'r'"

p = partial(foo, 0,x=4)
try:
    res1 = p(1,2,3,4,r=5)
    res2 = Exception()
except:
    res1,res2 = sys.exc_info()[:2]
try:
    re_1 = validate(p,1,2,3,4,r=5)
    re_2 = Exception()
except:
    re_1,re_2 = sys.exc_info()[:2]
#print(res2)
#print(re_2)
assert res1 == re_1
if hex(sys.hexversion).startswith('0x207'):
    assert str(res2) == str(TypeError("foo() got multiple values for keyword argument 'x'"))
elif hex(sys.hexversion) < '0x30300f0': # all versions slightly different
    assert str(res2)[:25] == str(re_2)[:25]
else:
    pass # python 3.3 gives "foo() got unexpected keyword argument 'r'"

p = partial(foo, 0,1,2,3,4,5,6,7)
try:
    res1 = p()
    res2 = Exception()
except:
    res1,res2 = sys.exc_info()[:2]
try:
    re_1 = validate(p)
    re_2 = Exception()
except:
    re_1,re_2 = sys.exc_info()[:2]
#print(res2)
#print(re_2)
assert res1 == re_1
if hex(sys.hexversion).startswith('0x207'):
    assert str(res2) == str(TypeError("foo() takes at most 5 arguments (8 given)"))
elif hex(sys.hexversion) < '0x30300f0': # all versions slightly different
    assert str(res2)[:20] == str(re_2)[:20]
else:
    pass # python 3.3 gives "foo() takes from 3 to 5 positional arguments but 8 were given"

p = partial(foo, 0,z=1,b=2)
try:
    res1 = p()
    res2 = Exception()
except:
    res1,res2 = sys.exc_info()[:2]
try:
    re_1 = validate(p)
    re_2 = Exception()
except:
    re_1,re_2 = sys.exc_info()[:2]
#print(res2)
#print(re_2)
assert res1 == re_1
if hex(sys.hexversion).startswith('0x207'):
    assert str(res2) == str(TypeError("foo() takes at least 3 arguments (3 given)"))
elif hex(sys.hexversion) < '0x30300f0': # all versions slightly different
    assert str(res2)[:21] == str(re_2)[:21]
else:
    pass # python 3.3 gives "foo() missing 1 required positional argument: 'y'"

p = partial(foo, 0,r=4)
try:
    res1 = p(1)
    res2 = Exception()
except:
    res1,res2 = sys.exc_info()[:2]
try:
    re_1 = validate(p,1)
    re_2 = Exception()
except:
    re_1,re_2 = sys.exc_info()[:2]
#print(res2)
#print(re_2)
assert res1 == re_1
if hex(sys.hexversion).startswith('0x207'):
    assert str(res2) == str(TypeError("foo() got an unexpected keyword argument 'r'"))
else: # all versions slightly different
    assert str(res2)[:24] == str(re_2)[:24]

p = partial(foo, 0,a=2)
try:
    res1 = p(1)
    res2 = Exception()
except:
    res1,res2 = sys.exc_info()[:2]
try:
    re_1 = validate(p,1)
    re_2 = Exception()
except:
    re_1,re_2 = sys.exc_info()[:2]
#print(res2)
#print(re_2)
assert res1 == re_1
if hex(sys.hexversion).startswith('0x207'):
    assert str(res2) == str(TypeError("foo() takes at least 3 arguments (3 given)"))
elif hex(sys.hexversion) < '0x30300f0': # all versions slightly different
    assert str(res2)[:21] == str(re_2)[:21]
else:
    pass # python 3.3 gives "foo() missing 1 required positional argument: 'z'"

assert validate(p,1,2) == None #XXX: better return ((1,2),{}) ?
'''
>>> p(1,2)
7
'''





#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 2013-2014 California Institute of Technology.
# License: 3-clause BSD.  The full license text is available at:
#  - http://trac.mystic.cacr.caltech.edu/project/pathos/browser/klepto/LICENSE
"""
The decorator should produce the behavior as displayed in the following:

>>> s = Spam()
>>> s.eggs()
new: (), {}
42
>>> s.eggs()
42
>>> s.eggs(1)
new: (1,), {}
64
>>> s.eggs(1)
64
>>> s.eggs(1, bar='spam')
new: (1,), {'bar': 'spam'}
78
>>> s2 = Spam()
>>> s2.eggs(1, bar='spam')
78
"""

from klepto.safe import inf_cache as memoized
#from klepto import inf_cache as memoized
from klepto.keymaps import picklemap
dumps = picklemap(flat=False, serializer='dill')

class Spam(object):
    """A simple class with a memoized method"""

    @memoized(keymap=dumps, ignore='self')
    def eggs(self, *args, **kwds):
       #print ('new:', args, kwds)
        from random import random
        return int(100 * random())

s = Spam()
assert s.eggs() == s.eggs()
assert s.eggs(1) == s.eggs(1)
s2 = Spam() 
assert s.eggs(1, bar='spam') == s2.eggs(1, bar='spam')
assert s.eggs.info().hit  == 3
assert s.eggs.info().miss == 3
assert s.eggs.info().load == 0

#print ('=' * 30)


# here caching saves time in a recursive function...
@memoized(keymap=dumps)
def fibonacci(n):
    "Return the nth fibonacci number."
   #print ('calculating %s' % n)
    if n in (0, 1):
        return n
    return fibonacci(n-1) + fibonacci(n-2)

fibonacci(7)
fibonacci(9)
fibonacci(3)
assert fibonacci.info().hit  == 9
assert fibonacci.info().miss == 10
assert fibonacci.info().load == 0

#print ('=' * 30)

try:
    from numpy import sum, asarray
    @memoized(keymap=dumps, tol=3)
    def add(*args):
       #print ('new:', args)
        return sum(args)

    assert add(1,2,3.0001)  == 6.0000999999999998
    assert add(1,2,3.00012) == 6.0000999999999998
    assert add(1,2,3.0234)  == 6.0234000000000005
    assert add(1,2,3.023)   == 6.0234000000000005
    assert add.info().hit  == 2
    assert add.info().miss == 2
    assert add.info().load == 0

    def cost(x,y):
       #print ('new: %s or %s' % (str(x), str(y)))
        x = asarray(x)
        y = asarray(y)
        return sum(x**2 - y**2)

    cost1 = memoized(keymap=dumps, tol=1)(cost)
    cost0 = memoized(keymap=dumps, tol=0)(cost)
    costD = memoized(keymap=dumps, tol=0, deep=True)(cost)

   #print ("rounding to one decimals...")
    cost1([1,2,3.1234], 3.9876)# == -32.94723372
    cost1([1,2,3.1234], 3.9876)# == -32.94723372
    cost1([1,2,3.1234], 3.6789)# == -25.84728807
    cost1([1,2,3.4321], 3.6789)# == -23.82360522
    assert cost1.info().hit  == 1
    assert cost1.info().miss == 3
    assert cost1.info().load == 0

   #print ("\nrerun the above with rounding to zero decimals...")
    cost0([1,2,3.1234], 3.9876)# == -32.94723372
    cost0([1,2,3.1234], 3.9876)# == -32.94723372
    cost0([1,2,3.1234], 3.6789)# == -32.94723372
    cost0([1,2,3.4321], 3.6789)# == -23.82360522
    assert cost0.info().hit  == 2
    assert cost0.info().miss == 2
    assert cost0.info().load == 0

   #print ("\nrerun again with deep rounding to zero decimals...")
    costD([1,2,3.1234], 3.9876)# == -32.94723372
    costD([1,2,3.1234], 3.9876)# == -32.94723372
    costD([1,2,3.1234], 3.6789)# == -32.94723372
    costD([1,2,3.4321], 3.6789)# == -32.94723372
    assert costD.info().hit  == 3
    assert costD.info().miss == 1
    assert costD.info().load == 0
   #print ("")
except ImportError:
    pass


import sys
import dill
from klepto.archives import cache, sql_archive, dict_archive
@memoized(cache=sql_archive())
def add(x,y):
    return x+y
add(1,2)
add(1,2)
add(1,3)
#print ("sql_cache = %s" % add.__cache__())
_key4 = '((), '+str({'y':3, 'x':1})+')'
_key3 = '((), '+str({'y':2, 'x':1})+')'
assert add.__cache__() == {_key4: 4, _key3: 3}

@memoized(cache=dict_archive(cached=False)) # use archive backend 'direcly'
def add(x,y):
    return x+y
add(1,2)
add(1,2)
add(1,3)
#print ("dict_cache = %s" % add.__cache__())
_key4 = '((), '+str({'y':3, 'x':1})+')'
_key3 = '((), '+str({'y':2, 'x':1})+')'
assert add.__cache__() == {_key4: 4, _key3: 3}

@memoized(cache=dict())
def add(x,y):
    return x+y
add(1,2)
add(1,2)
add(1,3)
#print ("dict_cache = %s" % add.__cache__())
assert add.__cache__() == {_key4: 4, _key3: 3}

@memoized(cache=add.__cache__())
def add(x,y):
    return x+y
add(1,2)
add(2,2)
#print ("re_dict_cache = %s" % add.__cache__())
_key2 = '((), '+str({'y':2, 'x':2})+')'
assert add.__cache__() == {_key4: 4, _key3: 3, _key2: 4}

@memoized(keymap=dumps)
def add(x,y):
    return x+y
add(1,2)
add(1,2)
add(1,3)
#print ("pickle_dict_cache = %s" % add.__cache__())
_key4 = eval(_key4)
_key3 = eval(_key3)
_pkey4 = dill.dumps(_key4)
_pkey3 = dill.dumps(_key3)
assert add.__cache__() == {_pkey4: 4, _pkey3: 3}


# EOF



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 2013-2014 California Institute of Technology.
# License: 3-clause BSD.  The full license text is available at:
#  - http://trac.mystic.cacr.caltech.edu/project/pathos/browser/klepto/LICENSE

from klepto.crypto import *
from klepto.tools import _b

assert string('1') == '1'
assert string('1', encoding='repr') == "'1'"

x = [1,2,3,'4',"'5'", min]
assert hash(x, 'sha1') == '3bdd73e79be4277dcb874d193b8dd08a46bc6885'
assert pickle(x) == string(x, 'repr')
assert string(x) == '[1, 2, 3, \'4\', "\'5\'", <built-in function min>]'
assert string(x, encoding='repr') == '[1, 2, 3, \'4\', "\'5\'", <built-in function min>]'
assert string(x, encoding='utf_8') == _b('[1, 2, 3, \'4\', "\'5\'", <built-in function min>]')
# some encodings 'missing' from klepto in python 3.x (due to bytes madness)
if 'unicode' in encodings():
    assert string(x, encoding='unicode') == unicode('[1, 2, 3, \'4\', "\'5\'", <built-in function min>]')
if 'zlib_codec' in encodings():
    assert string(x, encoding='zlib_codec') == 'x\x9c\x8b6\xd4Q0\xd2Q0\xd6QP7Q\xd7QPR7UW\xd2Q\xb0I*\xcd\xcc)\xd1\xcd\xccSH+\xcdK.\xc9\xcc\xcfS\xc8\xcd\xcc\xb3\x8b\x05\x00\xf6(\x0c\x9c'
if 'hex_codec' in encodings():
    assert string(x, encoding='hex_codec') == '5b312c20322c20332c202734272c2022273527222c203c6275696c742d696e2066756e6374696f6e206d696e3e5d'

from klepto.keymaps import *
s = stringmap()
assert s(x) == '([1, 2, 3, \'4\', "\'5\'", <built-in function min>],)'
s = stringmap(encoding='utf_8')
assert s(x) == _b('([1, 2, 3, \'4\', "\'5\'", <built-in function min>],)')
# some encodings 'missing' from klepto in python 3.x (due to bytes madness)
if 'unicode' in encodings():
    s = stringmap(encoding='unicode')
    assert s(x) == unicode('([1, 2, 3, \'4\', "\'5\'", <built-in function min>],)')
if 'zlib_codec' in encodings():
    s = stringmap(encoding='zlib_codec')
    assert s(x) == 'x\x9c\xd3\x886\xd4Q0\xd2Q0\xd6QP7Q\xd7QPR7UW\xd2Q\xb0I*\xcd\xcc)\xd1\xcd\xccSH+\xcdK.\xc9\xcc\xcfS\xc8\xcd\xcc\xb3\x8b\xd5\xd1\x04\x00\x17\x99\r\x19'
if 'hex_codec' in encodings():
    s = stringmap(encoding='hex_codec')
    assert s(x) == '285b312c20322c20332c202734272c2022273527222c203c6275696c742d696e2066756e6374696f6e206d696e3e5d2c29'



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 2013-2014 California Institute of Technology.
# License: 3-clause BSD.  The full license text is available at:
#  - http://trac.mystic.cacr.caltech.edu/project/pathos/browser/klepto/LICENSE

import dill
import klepto

@klepto.lru_cache()
def squared(x):
    return x**2

squared(2)
squared(4)
squared(6)

_s = dill.loads(dill.dumps(squared))
assert _s.lookup(4) == 16
assert squared.__cache__() == _s.__cache__()




# THIS FILE GENERATED FROM SETUP.PY
this_version = '0.1.1'
stable_version = '0.1.1'
readme = '''---------------------------------------------------
klepto: persistent caching to memory, disk, or database
---------------------------------------------------

Klepto extends python's 'lru_cache' to utilize different keymaps and
alternate caching algorithms, such as 'lfu_cache' and 'mru_cache'.
While caching is meant for fast access to saved results, klepto also
has archiving capabilities, for longer-term storage. Klepto uses a
simple dictionary-sytle interface for all caches and archives, and all
caches can be applied to any python function as a decorator. Keymaps
are algorithms for converting a function's input signature to a unique
dictionary, where the function's results are the dictionary value.
Thus for y = f(x), y will be stored in cache[x] (e.g. {x:y}).

Klepto provides both standard and 'safe' caching, where safe caches
are slower but can recover from hashing errors. Klepto is intended
to be used for distributed and parallel computing, where several of
the keymaps serialize the stored objects. Caches and archives are
intended to be read/write accessible from different threads and
processes. Klepto enables a user to decorate a function, save the
results to a file or database archive, close the interpreter,
start a new session, and reload the function and it's cache.

Klepto is part of pathos, a python framework for heterogenous computing.
Klepto is in the early development stages, and any user feedback is
highly appreciated. Contact Mike McKerns [mmckerns at caltech dot edu]
with comments, suggestions, and any bugs you may find. A list of known
issues is maintained at http://dev.danse.us/trac/pathos/query.


Major Features
==============

Klepto has standard and 'safe' variants of the following::

    - 'lfu_cache' - the least-frequently-used caching algorithm
    - 'lru_cache' - the least-recently-used caching algorithm
    - 'mru_cache' - the most-recently-used caching algorithm
    - 'rr_cache' - the random-replacement caching algorithm
    - 'no_cache' - a dummy caching interface to archiving
    - 'inf_cache' - an infinitely-growing cache

Klepto has the following archive types::

    - 'file_archive' - a dictionary-style interface to a file
    - 'dir_archive' - a dictionary-style interface to a folder of files
    - 'sqltable_archive' - a dictionary-style interface to a sql database table
    - 'sql_archive' - a dictionary-style interface to a sql database
    - 'dict_archive' - a dictionary with an archive interface
    - 'null_archive' - a dictionary-style interface to a dummy archive 

Klepto provides the following keymaps::

    - 'keymap' - keys are raw python objects
    - 'hashmap' - keys are a hash for the python object
    - 'stringmap' - keys are the python object cast as a string
    - 'picklemap' - keys are the serialized python object

Klepto also includes a few useful decorators providing::

    - simple, shallow, or deep rounding of function arguments
    - cryptographic key generation, with masking of selected arguments


Current Release
===============

This release version is klepto-0.1.1. You can download it here.
The latest released version of klepto is always available from::

    http://dev.danse.us/trac/pathos

Klepto is distributed under a 3-clause BSD license.


Development Release
===================

You can get the latest development release with all the shiny new features at::

    http://dev.danse.us/packages

or even better, fork us on our github mirror of the svn trunk::

    https://github.com/uqfoundation


Installation
============

Klepto is packaged to install from source, so you must
download the tarball, unzip, and run the installer::

    [download]
    $ tar -xvzf klepto-0.1.1.tgz
    $ cd klepto-0.1.1
    $ python setup py build
    $ python setup py install

You will be warned of any missing dependencies and/or settings
after you run the "build" step above. 

Alternately, klepto can be installed with easy_install or pip::

    [download]
    $ easy_install -f . klepto


Requirements
============

Klepto requires::

    - python2, version >= 2.5  *or*  python3, version >= 3.1
    - dill, version >= 0.2.1
    - pox, version >= 0.2

Optional requirements::

    - sqlalchemy, version >= 0.8.4
    - setuptools, version >= 0.6


Usage Notes
===========

Probably the best way to get started is to look at the tests
that are provide within klepto. See `klepto.tests` for a set of scripts
that test klepto's caching and archiving functionalities. Klepto's
source code is also generally well documented, so further questions may
be resolved by inspecting the code itself.


License
=======

Klepto is distributed under a 3-clause BSD license.

    >>> import klepto
    >>> print (klepto.license())


Citation
========

If you use klepto to do research that leads to publication,
we ask that you acknowledge use of klepto by citing the
following in your publication::

    Michael McKerns and Michael Aivazis,
    "pathos: a framework for heterogeneous computing", 2010- ;
    http://dev.danse.us/trac/pathos


More Information
================

Please see http://dev.danse.us/trac/pathos for further information.

'''
license = '''This software is part of the open-source mystic project at the California
Institute of Technology, and is available subject to the conditions and
terms laid out below. By downloading and using this software you are
agreeing to the following conditions.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met::

    - Redistribution of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.

    - Redistribution in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentations and/or other materials provided with the distribution.

    - Neither the name of the California Institute of Technology nor
      the names of its contributors may be used to endorse or promote
      products derived from this software without specific prior written
      permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

Copyright (c) 2014 California Institute of Technology. All rights reserved.

'''



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 2013-2014 California Institute of Technology.
# License: 3-clause BSD.  The full license text is available at:
#  - http://trac.mystic.cacr.caltech.edu/project/pathos/browser/klepto/LICENSE

from klepto.keymaps import *
from klepto.crypto import *
h = hashmap(algorithm='md5')
p = picklemap(serializer='dill')
hp = p + h
assert p(1) == pickle(1, serializer='dill')
assert h(1) == 'c4ca4238a0b923820dcc509a6f75849b'
if sys.version_info[0] == 2:
    assert hp(1) == 'ee7f32117eab97ec5460313282938c93'
else: #XXX: different, because 3.x returns b'' while 2.x returns ''
    assert hp(1) == 'a2ed37e4f2f0ccf8be170d8c31c711b2'
assert h(p(1)) == hp(1)
assert hp.inner(1) == p(1)
assert hp.outer(1) == h(1)
assert bool(h.inner) == False
assert bool(p.inner) == False
assert bool(hp.inner) == True
assert bool(h.outer) == False
assert bool(p.outer) == False
assert bool(hp.outer) == True


# EOF



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 2013-2014 California Institute of Technology.
# License: 3-clause BSD.  The full license text is available at:
#  - http://trac.mystic.cacr.caltech.edu/project/pathos/browser/klepto/LICENSE

from klepto.keymaps import *

args = (1,2); kwds = {"a":3, "b":4}

encode = keymap(typed=False, flat=True, sentinel=NOSENTINEL)
assert encode(*args, **kwds) == (1, 2, 'a', 3, 'b', 4)
encode = keymap(typed=False, flat=False, sentinel=NOSENTINEL)
assert encode(*args, **kwds) == (args, kwds)
encode = keymap(typed=True, flat=True, sentinel=NOSENTINEL)
assert encode(*args, **kwds) == (1, 2, 'a', 3, 'b', 4, type(1), type(2), type(3), type(4))
encode = keymap(typed=True, flat=False, sentinel=NOSENTINEL)
assert encode(*args, **kwds) == (args, kwds, (type(1), type(2)), (type(3), type(4)))

encode = hashmap(typed=False, flat=True, sentinel=NOSENTINEL)
assert encode(*args, **kwds) == hash((1, 2, 'a', 3, 'b', 4))
#encode = hashmap(typed=False, flat=False, sentinel=NOSENTINEL)
#assert encode(*args, **kwds) == TypeError("unhashable type: 'dict'")
encode = hashmap(typed=True, flat=True, sentinel=NOSENTINEL)
assert encode(*args, **kwds) == hash((1, 2, 'a', 3, 'b', 4, type(1), type(2), type(3), type(4)))
#encode = hashmap(typed=True, flat=False, sentinel=NOSENTINEL)
#assert encode(*args, **kwds) == TypeError("unhashable type: 'dict'")

encode = stringmap(typed=False, flat=True, sentinel=NOSENTINEL)
assert encode(*args, **kwds) == "(1, 2, 'a', 3, 'b', 4)"
encode = stringmap(typed=False, flat=False, sentinel=NOSENTINEL)
assert encode(*args, **kwds) == str((args, kwds))
encode = stringmap(typed=True, flat=True, sentinel=NOSENTINEL)
assert encode(*args, **kwds) == str( (1, 2, 'a', 3, 'b', 4, type(1), type(2), type(3), type(4)) )
encode = stringmap(typed=True, flat=False, sentinel=NOSENTINEL)
assert encode(*args, **kwds) == str( (args, kwds, (type(1), type(2)), (type(3), type(4))) )

from dill import dumps
encode = picklemap(typed=False, flat=True, serializer='dill')
assert encode(*args, **kwds) == dumps((1, 2, 'a', 3, 'b', 4))
encode = picklemap(typed=False, flat=False, serializer='dill')
assert encode(*args, **kwds) == dumps((args, kwds))
encode = picklemap(typed=True, flat=True, serializer='dill')
assert encode(*args, **kwds) == dumps( (1, 2, 'a', 3, 'b', 4, type(1), type(2), type(3), type(4)) )
encode = picklemap(typed=True, flat=False, serializer='dill')
assert encode(*args, **kwds) == dumps( (args, kwds, (type(1), type(2)), (type(3), type(4))) )




#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 2013-2014 California Institute of Technology.
# License: 3-clause BSD.  The full license text is available at:
#  - http://trac.mystic.cacr.caltech.edu/project/pathos/browser/klepto/LICENSE
"""
test speed and effectiveness of a selection of cache algorithms
"""

from klepto.archives import file_archive
from random import choice, seed

def _test_hits(algorithm, maxsize=20, keymap=None,
               rangelimit=5, tries=1000, archived=False):

    @algorithm(maxsize=maxsize, keymap=keymap)
    def f(x, y):
        return 3*x+y

    if archived:
        f.archive(file_archive('cache.pkl',cached=False))

    domain = list(range(rangelimit))
    domain += [float(i) for i in domain]
    for i in range(tries):
        r = f(choice(domain), choice(domain))

    f.dump()
   #print(f.info())
    return f.info()


if __name__ == '__main__':

    import os 
    import sys
    PY32 = ( hex(sys.hexversion) >= '0x30200f0' )
    from klepto import *
   #from klepto.safe import *
    seed(1234) # random seed

    caches = [rr_cache,mru_cache,lru_cache,lfu_cache,inf_cache,no_cache]

    # clean-up
    if os.path.exists('cache.pkl'): os.remove('cache.pkl')

   #print ("WITHOUT ARCHIVE")
    results = [_test_hits(cache, maxsize=100,
                          rangelimit=20, tries=100) for cache in caches]
    x = results[0]
    if PY32:
        assert (x.hit, x.miss, x.load, x.maxsize, x.size) == (16,84,0,100,84)
    else:
        assert (x.hit, x.miss, x.load, x.maxsize, x.size) == (13,87,0,100,87)
    x = results[1]
    if PY32:
        assert (x.hit, x.miss, x.load, x.maxsize, x.size) == (12,88,0,100,88)
    else:
        assert (x.hit, x.miss, x.load, x.maxsize, x.size) == (13,87,0,100,87)
    x = results[2]
    if PY32:
        assert (x.hit, x.miss, x.load, x.maxsize, x.size) == (11,89,0,100,89)
    else:
        assert (x.hit, x.miss, x.load, x.maxsize, x.size) == (10,90,0,100,90)
    x = results[3]
    if PY32:
        assert (x.hit, x.miss, x.load, x.maxsize, x.size) == (11,89,0,100,89)
    else:
        assert (x.hit, x.miss, x.load, x.maxsize, x.size) == (8,92,0,100,92)
    x = results[4]
    if PY32:
        assert (x.hit, x.miss, x.load, x.maxsize, x.size) == (11,89,0,None,89)
    else:
        assert (x.hit, x.miss, x.load, x.maxsize, x.size) == (9,91,0,None,91)
    x = results[5]
    assert (x.hit, x.miss, x.load, x.maxsize, x.size) == (0,100,0,0,0)
   #for cache in caches:
   #    msg = cache.__name__ + ":"
   #    msg += "%s" % str(_test_hits(cache, maxsize=100, 
   #                                 rangelimit=20, tries=100))
   #    print (msg)

   #print ("\nWITH ARCHIVE")
    results = [_test_hits(cache, maxsize=100, rangelimit=20,
                          tries=100, archived=True) for cache in caches]
    # clean-up
    if os.path.exists('cache.pkl'): os.remove('cache.pkl')

    x = results[0]
    if PY32:
        assert (x.hit, x.miss, x.load, x.maxsize, x.size) == (11,89,0,100,89)
    else:
        assert (x.hit, x.miss, x.load, x.maxsize, x.size) == (11,89,0,100,89)
    x = results[1]
    if PY32:
        assert (x.hit, x.miss, x.load, x.maxsize, x.size) == (10,66,24,100,90)
    else:
        assert (x.hit, x.miss, x.load, x.maxsize, x.size) == (11,68,21,100,89)
    x = results[2]
    if PY32:
        assert (x.hit, x.miss, x.load, x.maxsize, x.size) == (10,58,32,100,90)
    else:
        assert (x.hit, x.miss, x.load, x.maxsize, x.size) == (9,55,36,100,91)
    x = results[3]
    if PY32:
        assert (x.hit, x.miss, x.load, x.maxsize, x.size) == (10,37,53,100,90)
    else:
        assert (x.hit, x.miss, x.load, x.maxsize, x.size) == (12,40,48,100,88)
    x = results[4]
    if PY32:
        assert (x.hit, x.miss, x.load, x.maxsize, x.size) == (5,37,58,None,95)
    else:
        assert (x.hit, x.miss, x.load, x.maxsize, x.size) == (11,32,57,None,89)
    x = results[5]
    if PY32:
        assert (x.hit, x.miss, x.load, x.maxsize, x.size) == (0,25,75,0,0)
    else:
        assert (x.hit, x.miss, x.load, x.maxsize, x.size) == (0,25,75,0,0)
   #for cache in caches:
   #    msg = cache.__name__ + ":"
   #    msg += "%s" % str(_test_hits(cache, maxsize=100,
   #                                 rangelimit=20, tries=100, archived=True))
   #    print (msg)


# EOF



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 2013-2014 California Institute of Technology.
# License: 3-clause BSD.  The full license text is available at:
#  - http://trac.mystic.cacr.caltech.edu/project/pathos/browser/klepto/LICENSE

from __future__ import with_statement, absolute_import
import os

# set version numbers
stable_version = '0.1.1'
target_version = '0.1.1'
is_release = True

# check if easy_install is available
try:
#   import __force_distutils__ #XXX: uncomment to force use of distutills
    from setuptools import setup
    has_setuptools = True
except ImportError:
    from distutils.core import setup
    has_setuptools = False

# generate version number
if os.path.exists('klepto/info.py'):
    # is a source distribution, so use existing version
    os.chdir('klepto')
    with open('info.py','r') as f:
        f.readline() # header
        this_version = f.readline().split()[-1].strip("'")
    os.chdir('..')
elif stable_version == target_version:
    # we are building a stable release
    this_version = target_version
else:
    # we are building a distribution
    this_version = target_version + '.dev'
    if is_release:
        from datetime import date
        today = "".join(date.isoformat(date.today()).split('-'))
        this_version += "-" + today

# get the license info
with open('LICENSE') as file:
    license_text = file.read()

# generate the readme text
long_description = \
"""---------------------------------------------------
klepto: persistent caching to memory, disk, or database
---------------------------------------------------

Klepto extends python's 'lru_cache' to utilize different keymaps and
alternate caching algorithms, such as 'lfu_cache' and 'mru_cache'.
While caching is meant for fast access to saved results, klepto also
has archiving capabilities, for longer-term storage. Klepto uses a
simple dictionary-sytle interface for all caches and archives, and all
caches can be applied to any python function as a decorator. Keymaps
are algorithms for converting a function's input signature to a unique
dictionary, where the function's results are the dictionary value.
Thus for y = f(x), y will be stored in cache[x] (e.g. {x:y}).

Klepto provides both standard and 'safe' caching, where safe caches
are slower but can recover from hashing errors. Klepto is intended
to be used for distributed and parallel computing, where several of
the keymaps serialize the stored objects. Caches and archives are
intended to be read/write accessible from different threads and
processes. Klepto enables a user to decorate a function, save the
results to a file or database archive, close the interpreter,
start a new session, and reload the function and it's cache.

Klepto is part of pathos, a python framework for heterogenous computing.
Klepto is in the early development stages, and any user feedback is
highly appreciated. Contact Mike McKerns [mmckerns at caltech dot edu]
with comments, suggestions, and any bugs you may find. A list of known
issues is maintained at http://dev.danse.us/trac/pathos/query.


Major Features
==============

Klepto has standard and 'safe' variants of the following::

    - 'lfu_cache' - the least-frequently-used caching algorithm
    - 'lru_cache' - the least-recently-used caching algorithm
    - 'mru_cache' - the most-recently-used caching algorithm
    - 'rr_cache' - the random-replacement caching algorithm
    - 'no_cache' - a dummy caching interface to archiving
    - 'inf_cache' - an infinitely-growing cache

Klepto has the following archive types::

    - 'file_archive' - a dictionary-style interface to a file
    - 'dir_archive' - a dictionary-style interface to a folder of files
    - 'sqltable_archive' - a dictionary-style interface to a sql database table
    - 'sql_archive' - a dictionary-style interface to a sql database
    - 'dict_archive' - a dictionary with an archive interface
    - 'null_archive' - a dictionary-style interface to a dummy archive 

Klepto provides the following keymaps::

    - 'keymap' - keys are raw python objects
    - 'hashmap' - keys are a hash for the python object
    - 'stringmap' - keys are the python object cast as a string
    - 'picklemap' - keys are the serialized python object

Klepto also includes a few useful decorators providing::

    - simple, shallow, or deep rounding of function arguments
    - cryptographic key generation, with masking of selected arguments


Current Release
===============

This release version is klepto-%(relver)s. You can download it here.
The latest released version of klepto is always available from::

    http://dev.danse.us/trac/pathos

Klepto is distributed under a 3-clause BSD license.


Development Release
===================

You can get the latest development release with all the shiny new features at::

    http://dev.danse.us/packages

or even better, fork us on our github mirror of the svn trunk::

    https://github.com/uqfoundation


Installation
============

Klepto is packaged to install from source, so you must
download the tarball, unzip, and run the installer::

    [download]
    $ tar -xvzf klepto-%(thisver)s.tgz
    $ cd klepto-%(thisver)s
    $ python setup py build
    $ python setup py install

You will be warned of any missing dependencies and/or settings
after you run the "build" step above. 

Alternately, klepto can be installed with easy_install or pip::

    [download]
    $ easy_install -f . klepto


Requirements
============

Klepto requires::

    - python2, version >= 2.5  *or*  python3, version >= 3.1
    - dill, version >= 0.2.1
    - pox, version >= 0.2

Optional requirements::

    - sqlalchemy, version >= 0.8.4
    - setuptools, version >= 0.6


Usage Notes
===========

Probably the best way to get started is to look at the tests
that are provide within klepto. See `klepto.tests` for a set of scripts
that test klepto's caching and archiving functionalities. Klepto's
source code is also generally well documented, so further questions may
be resolved by inspecting the code itself.


License
=======

Klepto is distributed under a 3-clause BSD license.

    >>> import klepto
    >>> print (klepto.license())


Citation
========

If you use klepto to do research that leads to publication,
we ask that you acknowledge use of klepto by citing the
following in your publication::

    Michael McKerns and Michael Aivazis,
    "pathos: a framework for heterogeneous computing", 2010- ;
    http://dev.danse.us/trac/pathos


More Information
================

Please see http://dev.danse.us/trac/pathos for further information.

""" % {'relver' : stable_version, 'thisver' : this_version}

# write readme file
with open('README', 'w') as file:
    file.write(long_description)

# generate 'info' file contents
def write_info_py(filename='klepto/info.py'):
    contents = """# THIS FILE GENERATED FROM SETUP.PY
this_version = '%(this_version)s'
stable_version = '%(stable_version)s'
readme = '''%(long_description)s'''
license = '''%(license_text)s'''
"""
    with open(filename, 'w') as file:
        file.write(contents % {'this_version' : this_version,
                               'stable_version' : stable_version,
                               'long_description' : long_description,
                               'license_text' : license_text })
    return

# write info file
write_info_py()

# build the 'setup' call
setup_code = """
setup(name='klepto',
      version='%s',
      description='persistent caching to memory, disk, or database',
      long_description = '''%s''',
      author = 'Mike McKerns',
      maintainer = 'Mike McKerns',
      maintainer_email = 'mmckerns@caltech.edu',
      license = 'BSD',
      platforms = ['any'],
      url = 'http://www.cacr.caltech.edu/~mmckerns',
      classifiers = ('Intended Audience :: Developers',
                     'Programming Language :: Python',
                     'Topic :: Physics Programming'),

      packages = ['klepto'],
      package_dir = {'klepto':'klepto'},
""" % (target_version, long_description)

# add dependencies
dill_version = '>=0.2.1'
pox_version = '>=0.2'
sqlalchemy_version = '>=0.8.4'
import sys
if has_setuptools:
    setup_code += """
      zip_safe=False,
      dependency_links = ['http://dev.danse.us/packages/'],
      install_requires = ['dill%s','pox%s'],
""" % (dill_version, pox_version)

# add the scripts, and close 'setup' call
setup_code += """
      )
"""

# exec the 'setup' code
exec(setup_code)

# if dependencies are missing, print a warning
try:
    import dill
    import pox
    import sqlalchemy
except ImportError:
    print ("\n***********************************************************")
    print ("WARNING: One of the following dependencies is unresolved:")
    print ("    dill %s" % dill_version)
    print ("    pox %s" % pox_version)
    print ("    sqlalchemy %s" % sqlalchemy_version)
    print ("***********************************************************\n")


if __name__=='__main__':
    pass

# end of file
